import datetime
from typing import List, Optional
from croniter import croniter
from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass

from gcp_airflow_foundations.enums.source_type import SourceType
from gcp_airflow_foundations.base_class.landing_zone_config import LandingZoneConfig
from gcp_airflow_foundations.base_class.schema_options_config import SchemaOptionsConfig

partition_limit = 4000
ms_day = 86400000
expiration_options = {
    "@hourly":partition_limit/24,
    "@daily":partition_limit, 
    "@monthly":partition_limit*30
}


@dataclass
class SourceConfig:
    """
    Attributes:
        name : Name of source
        source_type : Source type selection. See SourceType class
        ingest_schedule : Ingestion schedule. Currently only supporting @hourly, @daily, @weekly, and @monthly
        gcp_project : GCP project ID
        dataset_data_name : Target dataset name
        connection : Aiflow GCP connection
        extra_options : GCP bucket and objects for source data if loading from GCS
        landing_zone_options : Staging dataset name
        acceptable_delay_minutes : Delay minutes limit
        notification_emails : Email address for notification emails
        owner : Airflow user owning the DAG
        partition_expiration: Expiration time for HDS Snapshot partitions in days.
        start_date : Start date for DAG
        start_date_tz : Timezone
        version : The Dag version. Can be incremented if logic changes
        sla_mins : SLA mins
    """
    name: str
    source_type: SourceType
    ingest_schedule: str
    gcp_project: str
    dataset_data_name: str
    dataset_hds_override: Optional[str]
    connection: str
    extra_options: dict
    landing_zone_options: LandingZoneConfig
    acceptable_delay_minutes: int
    notification_emails: List[str]
    owner: str
    partition_expiration: Optional[int]
    schema_options: SchemaOptionsConfig
    start_date: str
    start_date_tz: str = "EST"
    version: int = 1
    sla_mins: int = 900

    @validator("name")
    def valid_name(cls, v):
        assert v, "Name must not be empty"
        return v

    @validator("source_type")
    def valid_source_type(cls, v):
        assert v, "Source type must not be empty"
        return v

    @validator("ingest_schedule")
    def valid_ingest_schedule(cls, v):
        assert (v in ["@hourly", "@daily", "@weekly", "@monthly"] or croniter.is_valid(v)), \
            "invalid ingest schedule: see Airflow documentation for more details"
        return v

    @validator("landing_zone_options")
    def valid_landing_zone(cls, v):
        assert v.landing_zone_dataset, "Landing zone dataset name must not be empty"
        return v

    @validator("start_date")
    def valid_start_date(cls, v):
        assert datetime.datetime.strptime(v, "%Y-%m-%d"), \
            "The date format for Start Date should be YYYY-MM-DD"
        return v

    @validator("schema_options")
    def valid_schema_options(cls, v):
        assert v is not None, 'Schema options configuration must be provided'
        return v

    @root_validator(pre=True)
    def valid_partition_expiration(cls, values):
        if values['partition_expiration'] is not None:
            assert values['partition_expiration'] < expiration_options[values['ingest_schedule']], \
                f"The partition limit should be smaller than {expiration_options[values['ingest_schedule']]} days. It is currently set to {values['partition_expiration']}"
            values['partition_expiration'] = values['partition_expiration']*ms_day
        return values

    @root_validator(pre=True)
    def valid_hds_dataset(cls, values):
        if values['dataset_hds_override'] is None:
            values['dataset_hds_override'] = values['dataset_data_name']

        return values



from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass
from datetime import timedelta
from datetime import datetime
from typing import List

from gcp_airflow_foundations.base_class.source_table_config import SourceTableConfig
from gcp_airflow_foundations.base_class.source_config import SourceConfig
from gcp_airflow_foundations.enums.hds_table_type import HdsTableType
from gcp_airflow_foundations.enums.time_partitioning import TimePartitioning

import logging


partitioning_options = {
    "HOUR":"@hourly",
    "DAY":"@daily", 
    "MONTH":"@monthly"
}


@dataclass
class DataSourceTablesConfig:
    source: SourceConfig
    tables: List[SourceTableConfig]

    @root_validator(pre=True)
    def valid_partitioning(cls, values):
        ingest_schedule = values['source'].ingest_schedule

        for table in values['tables']:
            if (table.hds_config is not None) and (table.hds_config.hds_table_time_partitioning is not None) and (table.hds_config.hds_table_type == HdsTableType.SNAPSHOT):
                partitioning_time = table.hds_config.hds_table_time_partitioning.value
                assert partitioning_options[partitioning_time] == ingest_schedule, \
                    f"Invalid partitioning time selection for table `{table.table_name}` - partitioning time `{partitioning_time}` must match ingestion schedule `{ingest_schedule}`"
            
        return values

    def dagrun_timeout(self):
        return timedelta(minutes=self.source.acceptable_delay_minutes)

    def table_start_date(self, table_config: SourceTableConfig):
        return datetime.strptime(self.source.start_date, "%Y-%m-%d")

    def source_start_date(self):
        return datetime.strptime(self.source.start_date, "%Y-%m-%d")



from abc import ABC, abstractmethod

from gcp_airflow_foundations.common.gcp.source_schema.gcs import read_schema_from_gcs
from gcp_airflow_foundations.common.gcp.source_schema.bq import read_schema_from_bq


class SchemaSourceConfig(ABC):

    @abstractmethod
    def schema_method(self):
        pass

    @abstractmethod
    def schema_method_arguments(self, data_source, table_config):
        pass

class AutoSchemaSourceConfig(SchemaSourceConfig):

    def schema_method(self):
        return

    def schema_method_arguments(self, data_source, table_config):
        return


class GCSSchemaSourceConfig(SchemaSourceConfig):

    def schema_method(self):
        return read_schema_from_gcs

    def schema_method_arguments(self, data_source, table_config):
        return {
                'gcs_schema_object':data_source.schema_options.schema_object_template.format(table_name=table_config.table_name)
            }


class BQLandingZoneSchemaSourceConfig(SchemaSourceConfig):

    def schema_method(self):
        return read_schema_from_bq

    def schema_method_arguments(self, data_source, table_config):
        return {
                'dataset_id':data_source.landing_zone_options.landing_zone_dataset, 
                'table_id':table_config.landing_zone_table_name_override
            }



from dataclasses import fields
from urllib.parse import urlparse
from abc import ABC, abstractmethod, abstractproperty
import logging

from airflow.models.dag import DAG
from airflow.utils.task_group import TaskGroup
from airflow.operators.python_operator import PythonOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator

from gcp_airflow_foundations.base_class.salesforce_ingestion_config import SalesforceIngestionConfig

from gcp_airflow_foundations.operators.api.operators.sf_to_gcs_query_operator import SalesforceToGcsQueryOperator
from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder


class FTPtoBQDagBuilder(DagBuilder):
    """
    Builds DAGs to load files from FTP to a BigQuery Table.
    """

    def set_schema_method_type(self):
        self.schema_source_type = self.config.source.schema_options.schema_source_type     

    def get_bq_ingestion_task(self, table_config):
        taskgroup = TaskGroup(group_id="ftp_taskgroup")

        sensor_task = self.get_file_sensor(table_config)
        sensor_task.task_group = taskgroup

        file_ingestion_task = self.file_ingestion_task(table_config)
        file_ingestion_task.task_group = taskgroup

        load_to_landing_task = self.load_to_landing_task(table_config)
        load_to_landing_task.task_group = taskgroup

        sensor_task >> file_ingestion_task >> load_to_landing_task

        return taskgroup

    @abstractmethod
    def metadata_file_tasks(self, table_config):
        pass

    @abstractmethod
    def file_ingestion_task(self, table_config):
        """
        Implements an Airflow task to ingest the files from the FTP source into GCS (e.g. from an SFTP server or an AWS bucket)
        """
        pass

    @abstractmethod
    def file_sensor(self, table_config):
        """
        Returns an Airflow sensor that waits for the list of files specified by either
        (1) the metadata file provided, or
        (2) the files provided in the config
        """
        pass

    @abstractmethod
    def load_to_landing_task(self, table_config):
        pass

    def validate_extra_options(self):
        tables = self.config.tables


from dataclasses import fields
from urllib.parse import urlparse
from abc import ABC, abstractmethod, abstractproperty
import logging

from airflow.models.dag import DAG
from airflow.utils.task_group import TaskGroup
from airflow.operators.python_operator import PythonOperator
from airflow.operators.dummy import DummyOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator
from airflow.providers.google.cloud.sensors.gcs import GCSObjectExistenceSensor

from gcp_airflow_foundations.base_class.salesforce_ingestion_config import SalesforceIngestionConfig

from gcp_airflow_foundations.operators.api.operators.sf_to_gcs_query_operator import SalesforceToGcsQueryOperator
from gcp_airflow_foundations.operators.api.sensors.gcs_sensor import GCSObjectListExistenceSensor
from gcp_airflow_foundations.source_class.ftp_source import FTPtoBQDagBuilder
from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder


class GCSFiletoBQDagBuilder(FTPtoBQDagBuilder):
    """
    Builds DAGs to load files from GCS to a BigQuery Table.
    """

    def set_schema_method_type(self):
        self.schema_source_type = self.config.source.schema_options.schema_source_type     

    def get_bq_ingestion_task(self, table_config):

        taskgroup = TaskGroup(group_id="ftp_taskgroup")

        metadata_file_sensor = self.metadata_file_sensor(table_config)
        metadata_file_sensor.task_group

        sensor_task = self.get_file_sensor(table_config)
        sensor_task.task_group = taskgroup

        file_ingestion_task = self.file_ingestion_task(table_config)
        file_ingestion_task.task_group = taskgroup

        load_to_landing_task = self.load_to_landing_task(table_config)
        load_to_landing_task.task_group = taskgroup

        sensor_task >> file_ingestion_task >> load_to_landing_task

        return taskgroup

    def metadata_file_tasks(self, table_config):
        metadata_file = table_config.extra_options.get("gcs_table_config")["metadata_file"]
        bucket = self.config.source.extra_options["ftp_gcs_config"]["gcs_bucket"]

        file_sensor = GCSObjectExistenceSensor(
            task_id="wait_for_metadata_file",
            bucket=bucket,
            object=metadata_file
        )

    def file_ingestion_task(self, table_config):
        """
        No ingestion is needed - data is already in GCS, so return a dummy operator.
        """
        return DummyOperator(
            task_id="dummy_operator"
        )

    def file_sensor(self, table_config, files_to_wait_for):
        """
        Returns an Airflow sensor that waits for the list of files specified by either
        (1) the metadata file provided, or
        (2) the files provided in the config
        """
        bucket = self.config.source.extra_options["ftp_gcs_config"]["gcs_bucket"]

        return GCSObjectListExistenceSensor(
            task_id="wait_for_files",
            bucket=bucket,
            objects=files_to_wait_for
        )

    def load_to_landing_task(self, table_config, files_to_load):
        data_source = self.config.source

        # Parameters
        bucket = self.config.source.extra_options["gcs_bucket"]
        source_format = table_config.extra_options.get("gcs_config")["source_format"]
        gcp_project = data_source.gcp_project
        landing_dataset = data_source.landing_zone_options.landing_zone_dataset 
        destination_table = f"{gcp_project}:{landing_dataset}.{table_config.landing_zone_table_name_override}"

        return GCSToBigQueryOperator(
            task_id='import_csv_to_bq_landing',
            bucket=bucket,
            source_objects=files_to_load,
            source_format=source_format,
            destination_project_dataset_table=destination_table,
            write_disposition='WRITE_TRUNCATE',
            create_disposition='CREATE_IF_NEEDED',
            skip_leading_rows=1,
        )

    def validate_extra_options(self):
        tables = self.config.tables


from dacite import Config
from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass

from typing import List, Optional
import regex as re

from gcp_airflow_foundations.enums.schema_source_type import SchemaSourceType

supported_fields = ['table_name','date']

@dataclass
class SchemaOptionsConfig:
    schema_source_type: SchemaSourceType
    schema_object_template: Optional[str] # Supported tamplate fields: {table_name}, {date}
    
    @root_validator(pre=True)
    def valid_config(cls, values):
        if values['schema_source_type'] == SchemaSourceType.GCS:
            assert values['schema_object_template'] is not None, 'To read schema from GCS a file template must be provided'
        return values

    @validator("schema_object_template")
    def valid_schema_object_template(cls, v):
        if v is not None:
            fields = re.findall(r'\{(.*?)\}', v)
            assert all([i in supported_fields for i in fields]), f'The GCS schema file template contains invalid fields: {[i for i in fields if i not in supported_fields]}'
        return v



import logging

from airflow.models.dag import DAG
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator
from gcp_airflow_foundations.operators.gcp.gcs_to_bigquery import CustomGCSToBigQueryOperator

from gcp_airflow_foundations.source_class.source import DagBuilder

from gcp_airflow_foundations.enums.schema_source_type import SchemaSourceType


class GCStoBQDagBuilder(DagBuilder):
    """
    Builds DAGs to load a CSV file from GCS to a BigQuery Table.
    """
    source_type = "GCS"

    def set_schema_method_type(self):
        self.schema_source_type = self.config.source.schema_options.schema_source_type

    def get_bq_ingestion_task(self, table_config):
        # 1 Load CSV to BQ Landing Zone with schema auto-detection
        data_source = self.config.source

        schema_config = self.get_schema_method_class()

        load_to_bq_landing = CustomGCSToBigQueryOperator(
            task_id='import_csv_to_bq_landing',
            schema_config=schema_config,
            data_source=data_source,
            table_config=table_config,
            bucket=data_source.extra_options["gcs_bucket"],
            source_objects=data_source.extra_options["gcs_objects"],
            destination_project_dataset_table=f"{data_source.landing_zone_options.landing_zone_dataset}.{table_config.landing_zone_table_name_override}",
            write_disposition='WRITE_TRUNCATE',
            create_disposition='CREATE_IF_NEEDED',
            skip_leading_rows=1
        )

        return load_to_bq_landing
             
    def validate_extra_options(self):
        # Example of extra validation to do
        extra_options = self.config.source.extra_options

        # assert bucket and object/s are non-empty
        assert extra_options["gcs_bucket"]
        assert extra_options["gcs_objects"]


# -*- coding: utf-8 -*-

VERSION = (0, 2, 12)
TAG = "v{}.{}.{}".format(0, 2, 12)
PRERELEASE = None
REVISION = None


def generate_version(version, prerelease=None, revision=None):
    version_parts = [".".join(map(str, version))]
    if prerelease is not None:
        version_parts.append(f"-{prerelease}")
    if revision is not None:
        version_parts.append(f".{revision}")
    return "".join(version_parts)


__title__ = "gcp-airflow-foundations-dev"
__description__ = "Opinionated framework based on Airflow 2.0 for building pipelines to ingest data into a BigQuery data warehouse"
__url__ = "https://github.com/badal-io/gcp-airflow-foundations"
__download_url__ = f"https://github.com/badal-io/gcp-airflow-foundations/archive/refs/tags/{TAG}.tar.gz"
__version__ = generate_version(VERSION, prerelease=PRERELEASE, revision=REVISION)
__author__ = "Badal.io"
__author_email__ = "info@badal.io"
__license__ = "Apache 2.0"



from dacite import Config

from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from enum import Enum
import yaml
import dacite
import logging
import glob


def load_tables_config(config_path) -> DataSourceTablesConfig:
    table_config_file = open(config_path, "r").read()
    raw_config = yaml.load(table_config_file, Loader=yaml.FullLoader)
    config = dacite.from_dict(
        data_class=DataSourceTablesConfig, data=raw_config, config=Config(cast=[Enum])
    )
    logging.info(f"Config is {config}")
    return config


def load_tables_config_from_dir(dir_path):
    files = glob.glob(f"{dir_path}/*.yaml")
    logging.info(f"Loading config from  {files}")
    return list(map(load_tables_config, files))



from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional

@dataclass
class HdsTableMetadataConfig:
    """
    Attributes:
        eff_start_time_column_name : Insertion time column name
        eff_end_time_column_name: Expiry time column name
        partition_time_column_name: Partition time column name
        hash_column_name : Hash column name
    """
    eff_start_time_column_name: str = 'af_metadata_created_at'
    eff_end_time_column_name: str = 'af_metadata_expired_at'
    partition_time_column_name: str = 'partition_time'
    hash_column_name: str = 'af_metadata_row_hash'



from gcp_airflow_foundations.operators.gcp.hds.load_hds_taskgroup import hds_builder
from gcp_airflow_foundations.operators.gcp.ods.load_ods_taskgroup import ods_builder

from gcp_airflow_foundations.operators.gcp.schema_parsing.schema_parsing_operator import ParseSchema
from gcp_airflow_foundations.enums.ingestion_type import IngestionType
from gcp_airflow_foundations.enums.hds_table_type import HdsTableType

import logging


def load_builder(
    data_source,
    table_config,
    schema_config,
    preceding_task,
    dag):

    """
    Method for building all needed Task Groups based on the HdsTableConfig and OdsTableConfig options declared in the config files
    """

    project_id = data_source.gcp_project
    table_id = table_config.table_name
    dataset_id = data_source.dataset_data_name
    dataset_hds_id = data_source.dataset_hds_override
    landing_zone_dataset = data_source.landing_zone_options.landing_zone_dataset
    landing_zone_table_name_override = table_config.landing_zone_table_name_override
    surrogate_keys = table_config.surrogate_keys
    column_mapping = table_config.column_mapping
    ingestion_type = table_config.ingestion_type
    partition_expiration = data_source.partition_expiration
    ods_table_config = table_config.ods_config
    hds_table_config = table_config.hds_config

    if ingestion_type == IngestionType.INCREMENTAL:
        ods_table_config.table_id = f"{table_id}_ODS_Incremental"

    elif ingestion_type == IngestionType.FULL:
        ods_table_config.table_id = f"{table_id}_ODS_Full"

    parse_schema = ParseSchema(
        task_id="schema_parsing",
        schema_config=schema_config,
        column_mapping=column_mapping,
        data_source=data_source,
        table_config=table_config,
        dag=dag
    )

    ods_task_group= ods_builder(
        project_id=project_id,
        table_id=ods_table_config.table_id,
        dataset_id=dataset_id,
        landing_zone_dataset=landing_zone_dataset,
        landing_zone_table_name_override=landing_zone_table_name_override,
        surrogate_keys=surrogate_keys,
        column_mapping=column_mapping,
        ingestion_type=ingestion_type,
        ods_table_config=ods_table_config,
        dag=dag
    )
    
    hds_task_group = None
    if hds_table_config:
        if hds_table_config.hds_table_type == HdsTableType.SNAPSHOT:
            hds_table_config.table_id = f"{table_id}_HDS_Snapshot"
        
        elif hds_table_config.hds_table_type == HdsTableType.SCD2:
            hds_table_config.table_id = f"{table_id}_HDS_SCD2"

        hds_task_group = hds_builder(
            project_id=project_id,
            table_id=hds_table_config.table_id,
            dataset_id=dataset_hds_id,
            landing_zone_dataset=dataset_id,
            landing_zone_table_name_override=ods_table_config.table_id,
            surrogate_keys=surrogate_keys,
            column_mapping=column_mapping,
            ingestion_type=ingestion_type,
            partition_expiration=partition_expiration,
            hds_table_config=hds_table_config,
            dag=dag
        )

    if hds_task_group:
        preceding_task >> parse_schema >> ods_task_group >> hds_task_group
    else:
        preceding_task >> parse_schema >> ods_task_group



def parse_ods_schema(
    ods_metadata,
    schema_fields) -> list:

    ODS_EXTRA_FIELDS = [
        {
            "name": ods_metadata.ingestion_time_column_name,
            "type": "TIMESTAMP"
        },
        {
            "name": ods_metadata.update_time_column_name,
            "type": "TIMESTAMP"
        },
        {
            "name": ods_metadata.primary_key_hash_column_name,
            "type": "STRING"
        },
        {
            "name": ods_metadata.hash_column_name,
            "type": "STRING"
        }
    ]

    ods_schema_fields = []

    ods_schema_fields.extend(schema_fields)
    ods_schema_fields.extend(ODS_EXTRA_FIELDS)

    return ods_schema_fields    



from gcp_airflow_foundations.enums.source_type import SourceType

from gcp_airflow_foundations.source_class.gcs_source import GCStoBQDagBuilder
from gcp_airflow_foundations.source_class.twilio_source import TwilioToBQDagBuilder
from gcp_airflow_foundations.source_class.salesforce_source import SalesforcetoBQDagBuilder
from gcp_airflow_foundations.source_class.oracle_dataflow_source import OracleToBQDataflowDagBuilder
from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig

import logging


def get_dag_builder(source: SourceType, default_task_args: dict, config: DataSourceTablesConfig):
    logging.info(f"Get DAG builder for {source}")
    if source == SourceType.GCS:
        logging.info("Selecting GCS builder")
        return GCStoBQDagBuilder(default_task_args, config)
    elif source == SourceType.TWILIO:
        logging.info("Selecting Twilio builder")
        return TwilioToBQDagBuilder(default_task_args, config)
    elif source == SourceType.SALESFORCE:
        logging.info("Selecting Salesforce builder")
        return SalesforcetoBQDagBuilder(default_task_args, config)
    elif source == SourceType.ORACLE:
        logging.info("Selecting Oracle builder")
        return OracleToBQDataflowDagBuilder(default_task_args, config)
    else:
        logging.info(f"Invalid source {source}")
        raise Exception(f"Invalid source {source}")


import datetime

from airflow.models.dag import DAG
from airflow.models import Variable
from gcp_airflow_foundations.base_class.utils import load_tables_config_from_dir

from gcp_airflow_foundations.source_class import get_dag_builder

import logging


class DagParser:
    """ Parsed Airflow DAGs from the user-provided YAML configuration files """
    def __init__(self):
        self.conf_location = Variable.get("CONFIG_FILE_LOCATION", "/opt/airflow/gcp_airflow_foundations/examples/config")
        self.max_task_retries = Variable.get("max_task_retries", 3)

    def parse_dags(self):
        logging.info(f"Loading config from dir: {self.conf_location}")

        configs = load_tables_config_from_dir(self.conf_location)

        parsed_dags = {}
        for config in configs:
            logging.info(f"StartDate for {config.source.name}: {config.source_start_date()}")

            default_task_args = {
                "owner": config.source.owner,
                "retries": self.max_task_retries,
                "retry_exponential_backoff": True,
                "retry_delay": datetime.timedelta(seconds=300),
                "project_id": config.source.gcp_project,
                "start_date": config.source_start_date(),
                "email": config.source.notification_emails,
                "email_on_failure": True,
                "email_on_retry": False,
                "depends_on_past": False
            }

            # Old way of adding DAGs
            builder = get_dag_builder(config.source.source_type, default_task_args, config)
            dags = builder.build_dags()

            for dag in dags:
                parsed_dags[f"dags:source:{config.source.name}.{dag.dag_id}"] = dag

            """
            # new way
            builder = None
            for dag_builder in DagBuilder.sources:
                # if matching subclass of DagBuilder exists, then use it
                if config.source.source_type == dag_builder.source_type:
                    source_name = dag_builder.__class__.__name__
                    # pick out the right source
                    builder = globals()[source_name](default_task_args, config)
                    dags = builder.build_dags()
            
            if dags:
                parsed_dags[f"dags:source:{config.source.name}.{dag.dag_id}"] = dag
            """
            
        return parsed_dags



from typing import Optional

from airflow.models import BaseOperator, BaseOperatorLink
from airflow.contrib.operators.bigquery_operator import (
    BigQueryOperator,
    BigQueryCreateEmptyTableOperator,
)

from airflow.utils.decorators import apply_defaults
from airflow.contrib.hooks.bigquery_hook import BigQueryHook

from airflow.exceptions import AirflowException

import logging

from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryCreateEmptyTableOperator
)


class CustomBigQueryCreateEmptyTableOperator(BigQueryCreateEmptyTableOperator):
    @apply_defaults
    def __init__(
        self,
        *,
        project_id: str,
        dataset_id: str,
        table_id: str,
        time_partitioning: Optional[dict] = None,
        schema_task_id: Optional[str] = "schema_parsing",
        gcp_conn_id: str = "google_cloud_default",
        **kwargs,
    ) -> None:
        super(CustomBigQueryCreateEmptyTableOperator, self).__init__(
            project_id=project_id,
            dataset_id=dataset_id,
            table_id=table_id,
            table_resource=None,
            exists_ok=True,
            **kwargs,
        )

        self.table = table_id
        self.time_partitioning = time_partitioning
        self.schema_task_id = schema_task_id

    def pre_execute(self, context) -> None:
        schema_fields = self.xcom_pull(context=context, task_ids=self.schema_task_id)[self.table_id]

        self.table_resource={
                "schema":{'fields': schema_fields},
                "timePartitioning":self.time_partitioning,
                "encryptionConfiguration":None,
                "labels":None
        }


import json
import logging

from airflow.contrib.hooks.bigquery_hook import BigQueryHook

def read_schema_from_bq(
    dataset_id,
    table_id,
    google_cloud_storage_conn_id='google_cloud_default',
    bigquery_conn_id='google_cloud_default',
    **kwargs) -> list:
    """
        Helper method to load table schema from the staging table
    """

    bq_hook = BigQueryHook(bigquery_conn_id=bigquery_conn_id, delegate_to=None)

    schema = bq_hook.get_schema(
        dataset_id=dataset_id,
        table_id=table_id
    )

    return schema['fields']



from typing import Optional
from datetime import datetime

from airflow.models import BaseOperator, BaseOperatorLink
from airflow.contrib.operators.bigquery_operator import (
    BigQueryOperator,
    BigQueryCreateEmptyTableOperator,
)

from airflow.utils.decorators import apply_defaults
from airflow.contrib.hooks.bigquery_hook import BigQueryHook

from airflow.exceptions import AirflowException

import logging

from gcp_airflow_foundations.operators.gcp.hds.hds_sql_upsert_helpers import SqlHelperHDS
from gcp_airflow_foundations.enums.hds_table_type import HdsTableType
from gcp_airflow_foundations.base_class.hds_table_config import HdsTableConfig
from gcp_airflow_foundations.enums.ingestion_type import IngestionType


class MergeBigQueryHDS(BigQueryOperator):
    """
    Merges data into a BigQuery HDS table.
    
    :param project_id: GCP project ID  
    :type project_id: str
    :param stg_table_name: Source table name
    :type stg_table_name: str
    :param data_table_name: Target table name
    :type data_table_name: str
    :param stg_dataset_name: Source dataset name
    :type stg_dataset_name: str
    :param data_dataset_name: Target dataset name
    :type data_dataset_name: str
    :param surrogate_keys: Surrogate keys of the target table
    :type surrogate_keys: list
    :param delegate_to: The account to impersonate using domain-wide delegation of authority, if any
    :type delegate_to: str
    :param gcp_conn_id: Airflow GCP connection ID
    :type gcp_conn_id: str
    :param column_mapping: Column mapping
    :type column_mapping: dict
    :param ingestion_type: Source table ingestion time (Full or Incremental)
    :type ingestion_type: IngestionType
    :param hds_table_config: User-provided HDS configuration options
    :type hds_table_config: HdsTableConfig
    """
    
    template_fields = (
        "stg_table_name",
        "data_table_name",
        "stg_dataset_name"
    )

    @apply_defaults
    def __init__(
        self,
        *,
        project_id: str,
        stg_table_name: str,
        data_table_name: str,
        stg_dataset_name: str,
        data_dataset_name: str,
        surrogate_keys: [str],
        columns: Optional[list] = None,
        delegate_to: Optional[str] = None,
        gcp_conn_id: str = "google_cloud_default",
        column_mapping: dict,
        ingestion_type: IngestionType,
        hds_table_config: HdsTableConfig,
        **kwargs,
    ) -> None:
        super(MergeBigQueryHDS, self).__init__(
            delegate_to=delegate_to,
            gcp_conn_id=gcp_conn_id,
            use_legacy_sql=False,
            write_disposition="WRITE_APPEND",
            create_disposition="CREATE_NEVER",
            destination_dataset_table=None,
            sql="",
            **kwargs,
        )
        self.project_id = project_id
        self.stg_table_name = stg_table_name
        self.data_table_name = data_table_name
        self.stg_dataset_name = stg_dataset_name
        self.data_dataset_name = data_dataset_name
        self.surrogate_keys = surrogate_keys
        self.gcp_conn_id = gcp_conn_id
        self.delegate_to = delegate_to
        self.columns = columns
        self.column_mapping = column_mapping
        self.ingestion_type = ingestion_type
        self.hds_table_config = hds_table_config

    def pre_execute(self, context) -> None:
        if not self.columns:
            staging_columns = self.xcom_pull(context=context, task_ids="schema_parsing")['source_table_columns']
        else:
            staging_columns = self.columns

        if self.column_mapping:
            for i in staging_columns:
                if i not in self.column_mapping:
                    self.column_mapping[i] = i
            
            self.surrogate_keys  = [self.column_mapping[i] for i in self.surrogate_keys]
            source_columns = [self.column_mapping[i] for i in staging_columns]
            self.column_mapping = {self.column_mapping[i]:self.column_mapping[i] for i in self.column_mapping}

        else:
            source_columns = staging_columns
            self.column_mapping = {i:i for i in staging_columns}
                    
        self.log.info(
            f"Execute BigQueryMergeTableOperator {self.stg_table_name}, {self.data_table_name}"
        )

        sql = ""

        sql_helper = SqlHelperHDS(
            source_dataset=self.stg_dataset_name,
            target_dataset=self.data_dataset_name,
            source=self.stg_table_name,
            target=self.data_table_name,
            columns=source_columns,
            surrogate_keys=self.surrogate_keys,
            column_mapping=self.column_mapping,
            hds_metadata=self.hds_table_config.hds_metadata
        )

        if self.hds_table_config.hds_table_type == HdsTableType.SNAPSHOT:
            partitioning_dimension = self.hds_table_config.hds_table_time_partitioning.value
            sql_helper.time_partitioning = partitioning_dimension
            sql = sql_helper.create_snapshot_sql_with_hash()

            now = datetime.now()
            if partitioning_dimension == "HOUR":
                partition_id = now.strftime("%Y%m%d%H")
            elif partitioning_dimension == "DAY":
                partition_id = now.strftime("%Y%m%d")
            elif partitioning_dimension == "MONTH":
                partition_id = now.strftime("%Y%m")
            else:
                raise AirflowException(f"Could not determine partition ID format from `{partitioning_dimension}`")

            self.write_disposition = "WRITE_TRUNCATE"
            self.create_disposition = "CREATE_IF_NEEDED"
            self.destination_dataset_table = f"{self.data_dataset_name}.{self.data_table_name}${partition_id}"

        elif self.hds_table_config.hds_table_type == HdsTableType.SCD2:
            sql = sql_helper.create_scd2_sql_with_hash(
                ingestion_type=self.ingestion_type
            )

        else:
            raise AirflowException("Invalid HDS table type", self.hds_table_config.hds_table_type)

        logging.info(f"Executing sql: {sql}")

        self.sql = sql